'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const http = require('http');
const https = require('https');
const buffer = require('buffer');
const events = require('events');
const required = require('requires-port');
const stream$2 = require('stream');
const followRedirects = require('follow-redirects');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e["default"] : e; }

const http__default = /*#__PURE__*/_interopDefaultLegacy(http);
const https__default = /*#__PURE__*/_interopDefaultLegacy(https);
const required__default = /*#__PURE__*/_interopDefaultLegacy(required);
const followRedirects__default = /*#__PURE__*/_interopDefaultLegacy(followRedirects);

const upgradeHeader = /(^|,)\s*upgrade\s*($|,)/i;
const isSSL = /^https|wss/;
function setupOutgoing(outgoing, options, req, forward) {
  const target = options[forward || "target"];
  if (typeof target === "object") {
    if (!target.searchParams) {
      target.searchParams = new URLSearchParams();
    }
    if (target.path) {
      target.pathname = target.path;
      delete target.path;
    }
  }
  const sslEnabled = isSSL.test(target.protocol);
  outgoing.port = target.port || (sslEnabled ? 443 : 80);
  for (const opt of ["host", "hostname", "socketPath", "pfx", "key", "passphrase", "cert", "ca", "ciphers", "secureProtocol", "servername"]) {
    outgoing[opt] = target[opt];
  }
  outgoing.method = options.method || req.method;
  outgoing.headers = Object.assign({}, req.headers);
  if (options.headers) {
    Object.assign(outgoing.headers, options.headers);
  }
  if (options.auth) {
    outgoing.auth = options.auth;
  }
  if (options.ca)
    outgoing.ca = options.ca;
  if (sslEnabled) {
    outgoing.rejectUnauthorized = typeof options.secure === "undefined" ? true : options.secure;
  }
  outgoing.agent = options.agent || false;
  outgoing.localAddress = options.localAddress;
  if (!outgoing.agent) {
    outgoing.headers = outgoing.headers || {};
    if (typeof outgoing.headers.connection !== "string" || !upgradeHeader.test(outgoing.headers.connection)) {
      outgoing.headers.connection = "close";
    }
  }
  if (options.followRedirects && typeof options.followRedirects === "object") {
    outgoing.maxRedirects = options.followRedirects.maxRedirects;
    outgoing.maxBodyLength = options.followRedirects.maxBodyLength;
    outgoing.agents = options.followRedirects.agents;
    outgoing.beforeRedirect = options.followRedirects.beforeRedirect;
    outgoing.trackRedirects = options.followRedirects.trackRedirects;
  }
  const targetPath = target && options.prependPath !== false ? target.pathname || "" : "";
  const reqUrl = new URL(req.url, "http://example.com");
  for (const entry of target.searchParams.entries()) {
    reqUrl.searchParams.set(entry[0], entry[1]);
  }
  const params = reqUrl.search;
  let outgoingPath = !options.toProxy ? reqUrl.pathname || "" : req.url;
  outgoingPath = !options.ignorePath ? outgoingPath : "";
  outgoing.path = [targetPath, outgoingPath].filter(Boolean).join("/").replace(/\/+/g, "/") + params;
  if (options.changeOrigin) {
    outgoing.headers.host = required__default(outgoing.port, target.protocol) && !hasPort(outgoing.host?.toString()) ? outgoing.host + ":" + outgoing.port : outgoing.host;
  }
  return outgoing;
}
function setupSocket(socket) {
  socket.setTimeout(0);
  socket.setNoDelay(true);
  socket.setKeepAlive(true, 0);
  return socket;
}
function getPort(req) {
  const res = req.headers.host ? req.headers.host.match(/:(\d+)/) : "";
  return res ? res[1] : hasEncryptedConnection(req) ? "443" : "80";
}
function hasEncryptedConnection(req) {
  return Boolean(req.connection?.encrypted || req.socket?.encrypted || req.connection?.pair || req.socket?.pair);
}
function rewriteCookieProperty(header, config, property) {
  if (Array.isArray(header)) {
    return header.map(function(headerElement) {
      return rewriteCookieProperty(headerElement, config, property);
    });
  }
  return header.replace(new RegExp("(;\\s*" + property + "=)([^;]+)", "i"), function(match, prefix, previousValue) {
    let newValue;
    if (previousValue in config) {
      newValue = config[previousValue];
    } else if ("*" in config) {
      newValue = config["*"];
    } else {
      return match;
    }
    if (newValue) {
      return prefix + newValue;
    } else {
      return "";
    }
  });
}
function removeCookieProperty(header, property) {
  if (Array.isArray(header)) {
    return header.map(function(headerElement) {
      return removeCookieProperty(headerElement, property);
    });
  }
  return header.replace(new RegExp(";\\s*" + property + "[^;]*", "i"), "");
}
function mergeSetCookie(setCookie, upstreamSetCookie) {
  let existingCookieArray = setCookie || [], upstreamCookieArray = upstreamSetCookie || [];
  if (!Array.isArray(existingCookieArray)) {
    existingCookieArray = [existingCookieArray];
  }
  if (!Array.isArray(upstreamCookieArray)) {
    upstreamCookieArray = [upstreamCookieArray];
  }
  return existingCookieArray.concat(upstreamCookieArray);
}
function hasPort(host) {
  return !!~host.indexOf(":");
}

const redirectRegex = /^201|30(1|2|7|8)$/;
function removeChunked(req, res, proxyRes) {
  if (req.httpVersion === "1.0" || proxyRes.statusCode === 204 || proxyRes.statusCode === 304) {
    delete proxyRes.headers["transfer-encoding"];
  }
}
function setConnection(req, res, proxyRes) {
  if (req.httpVersion === "1.0") {
    proxyRes.headers.connection = req.headers.connection || "close";
  } else if (req.httpVersion !== "2.0" && !proxyRes.headers.connection) {
    proxyRes.headers.connection = req.headers.connection || "keep-alive";
  }
}
function setRedirectHostRewrite(req, res, proxyRes, options) {
  if ((options.hostRewrite || options.autoRewrite || options.protocolRewrite) && proxyRes.headers["location"] && redirectRegex.test(String(proxyRes.statusCode))) {
    const targetStr = typeof options.target === "string" ? options.target : options.target?.href;
    const target = new URL(targetStr);
    const u = new URL(proxyRes.headers["location"]);
    if (target.host != u.host) {
      return;
    }
    if (options.hostRewrite) {
      u.host = options.hostRewrite;
    } else if (options.autoRewrite) {
      u.host = req.headers["host"];
    }
    if (options.protocolRewrite) {
      u.protocol = options.protocolRewrite;
    }
    proxyRes.headers["location"] = u.toString();
  }
}
function writeHeaders(req, res, proxyRes, options) {
  const preserveHeaderKeyCase = options.preserveHeaderKeyCase;
  const mergeCookiesConfig = options.mergeCookies;
  const cookieRemoveSecure = options.cookieRemoveSecure;
  let rewriteCookieDomainConfig = options.cookieDomainRewrite;
  let rewriteCookiePathConfig = options.cookiePathRewrite;
  let rawHeaderKeyMap;
  const setHeader = function(key, header) {
    if (rewriteCookieDomainConfig && key.toLowerCase() === "set-cookie") {
      header = rewriteCookieProperty(header, rewriteCookieDomainConfig, "domain");
    }
    if (rewriteCookiePathConfig && key.toLowerCase() === "set-cookie") {
      header = rewriteCookieProperty(header, rewriteCookiePathConfig, "path");
    }
    if (mergeCookiesConfig && key.toLowerCase() === "set-cookie") {
      header = mergeSetCookie(res.getHeader("set-cookie"), header);
    }
    if (cookieRemoveSecure && key.toLowerCase() === "set-cookie") {
      header = removeCookieProperty(header, "secure");
    }
    try {
      res.setHeader(String(key).trim(), header);
    } catch (error) {
      console.warn(error, key, header);
    }
  };
  if (typeof rewriteCookieDomainConfig === "string") {
    rewriteCookieDomainConfig = { "*": rewriteCookieDomainConfig };
  }
  if (typeof rewriteCookiePathConfig === "string") {
    rewriteCookiePathConfig = { "*": rewriteCookiePathConfig };
  }
  if (preserveHeaderKeyCase && proxyRes.rawHeaders != void 0) {
    rawHeaderKeyMap = {};
    for (let i = 0; i < proxyRes.rawHeaders.length; i += 2) {
      let key = proxyRes.rawHeaders[i];
      rawHeaderKeyMap[key.toLowerCase()] = key;
    }
  }
  Object.keys(proxyRes.headers).forEach(function(key) {
    const header = proxyRes.headers[key];
    if (preserveHeaderKeyCase && rawHeaderKeyMap) {
      key = rawHeaderKeyMap[key] || key;
    }
    setHeader(key, header);
  });
}
function attachOutgoingHeaders(req, res, proxyRes, options) {
  if (options.outgoingHeaders != null) {
    Object.keys(options.outgoingHeaders).forEach(function(header) {
      res.setHeader(header, options.outgoingHeaders[header]);
    });
  }
}
function writeStatusCode(req, res, proxyRes) {
  if (proxyRes.statusMessage) {
    res.statusCode = proxyRes.statusCode;
    res.statusMessage = proxyRes.statusMessage;
  } else {
    res.statusCode = proxyRes.statusCode;
  }
}

const webOutgoing = {
    __proto__: null,
    removeChunked: removeChunked,
    setConnection: setConnection,
    setRedirectHostRewrite: setRedirectHostRewrite,
    writeHeaders: writeHeaders,
    attachOutgoingHeaders: attachOutgoingHeaders,
    writeStatusCode: writeStatusCode
};

const passes = Object.keys(webOutgoing).map((pass) => webOutgoing[pass]);
const nativeAgents = { http: http__default, https: https__default };
function deleteLength(req, res, options) {
  if ((req.method === "DELETE" || req.method === "OPTIONS") && !req.headers["content-length"]) {
    req.headers["content-length"] = "0";
    delete req.headers["transfer-encoding"];
  }
}
function timeout(req, res, options) {
  let timeoutValue = options.timeout ? options.timeout : 0;
  req.socket.setTimeout(timeoutValue);
}
function XHeaders$1(req, res, options) {
  if (!options.xfwd)
    return;
  let encrypted = req.isSpdy || hasEncryptedConnection(req);
  let values = {
    for: req.connection?.remoteAddress || req.socket?.remoteAddress,
    port: getPort(req),
    proto: encrypted ? "https" : "http"
  };
  for (let header of ["for", "port", "proto"]) {
    const headerName = "x-forwarded-" + header;
    if (!req.headers[headerName]) {
      req.headers[headerName] = values[header];
    }
  }
  req.headers["x-forwarded-host"] = req.headers["x-forwarded-host"] || req.headers["host"] || "";
}
function stream$1(req, res, options, head, server, callback) {
  server.emit("start", req, res, options.target || options.forward);
  const agents = options.followRedirects ? followRedirects__default : nativeAgents;
  const http = agents.http;
  const https = agents.https;
  if (options.forward) {
    const forwardReq = (isSSL.test(options.forward.protocol) ? https : http).request(setupOutgoing(options.ssl || {}, options, req, "forward"));
    const forwardError = createErrorHandler(forwardReq, options.forward);
    req.on("error", forwardError);
    forwardReq.on("error", forwardError);
    stream$2.pipeline(options.buffer || req, forwardReq, () => {
    });
    if (!options.target) {
      return res.end();
    }
  }
  const proxyReq = (isSSL.test(options.target.protocol) ? https : http).request(setupOutgoing(options.ssl || {}, options, req));
  proxyReq.on("socket", function(socket) {
    if (server && !proxyReq.getHeader("expect")) {
      server.emit("proxyReq", proxyReq, req, res, options);
    }
  });
  if (options.proxyTimeout) {
    proxyReq.setTimeout(options.proxyTimeout, function() {
      proxyReq.destroy();
    });
  }
  req.on("aborted", function() {
    proxyReq.destroy();
  });
  res.on("close", function() {
    if (res.destroyed) {
      proxyReq.destroy();
    }
  });
  const proxyError = createErrorHandler(proxyReq, options.target);
  proxyReq.on("error", proxyError);
  req.on("error", proxyError);
  function createErrorHandler(proxyReq2, url) {
    return function proxyError2(err) {
      if ((req.aborted || req.socket.destroyed) && err.code === "ECONNRESET") {
        server.emit("econnreset", err, req, res, url);
        proxyReq2.destroy();
        return;
      }
      if (callback) {
        callback(err, req, res, url);
      } else {
        server.emit("error", err, req, res, url);
      }
    };
  }
  stream$2.pipeline(options.buffer || req, proxyReq, () => {
  });
  proxyReq.on("response", function(proxyRes) {
    if (server) {
      server.emit("proxyRes", proxyRes, req, res, options);
    }
    const selfHandle = typeof options.selfHandleResponse === "function" ? options.selfHandleResponse(proxyRes, req, res) : options.selfHandleResponse;
    if (!res.headersSent && (!selfHandle || options.forcePasses)) {
      for (let i = 0; i < passes.length; i++) {
        if (typeof passes[i] === "function" && passes[i](req, res, proxyRes, options)) {
          break;
        }
      }
    }
    if (!res.finished) {
      proxyRes.on("end", function() {
        if (server) {
          server.emit("end", req, res, proxyRes);
        }
      });
      if (!selfHandle) {
        stream$2.pipeline(proxyRes, res, () => {
        });
      }
    } else {
      if (server) {
        server.emit("end", req, res, proxyRes);
      }
    }
  });
}

const webIncoming = {
    __proto__: null,
    deleteLength: deleteLength,
    timeout: timeout,
    XHeaders: XHeaders$1,
    stream: stream$1
};

function checkMethodAndHeader(req, socket) {
  if (req.method !== "GET" || !req.headers.upgrade) {
    socket.destroy();
    return true;
  }
  if (req.headers.upgrade.toLowerCase() !== "websocket") {
    socket.destroy();
    return true;
  }
}
function XHeaders(req, socket, options) {
  if (!options.xfwd)
    return;
  let values = {
    for: req.connection.remoteAddress || req.socket.remoteAddress,
    port: getPort(req),
    proto: hasEncryptedConnection(req) ? "wss" : "ws"
  };
  ["for", "port", "proto"].forEach(function(header) {
    req.headers["x-forwarded-" + header] = (req.headers["x-forwarded-" + header] || "") + (req.headers["x-forwarded-" + header] ? "," : "") + values[header];
  });
}
async function stream(req, socket, options, head, server, callback) {
  let createHttpHeader = function(line, headers) {
    return Object.keys(headers).reduce(function(head2, key) {
      let value = headers[key];
      if (!Array.isArray(value)) {
        head2.push(key + ": " + value);
        return head2;
      }
      for (let i = 0; i < value.length; i++) {
        head2.push(key + ": " + value[i]);
      }
      return head2;
    }, [line]).join("\r\n") + "\r\n\r\n";
  };
  setupSocket(socket);
  if (head && head.length) {
    socket.unshift(head);
  }
  let proxyReq = (isSSL.test(options.target.protocol) ? https__default : http__default).request(setupOutgoing(options.ssl || {}, options, req));
  proxyReq.on("error", onOutgoingError);
  proxyReq.on("response", function(res) {
    if (!res.upgrade && !socket.destroyed) {
      socket.write(createHttpHeader("HTTP/" + res.httpVersion + " " + res.statusCode + " " + res.statusMessage, res.headers));
      res.pipe(socket).on("error", onOutgoingError);
    }
  });
  proxyReq.on("upgrade", function(proxyRes, proxySocket, proxyHead) {
    proxySocket.on("error", onOutgoingError);
    proxySocket.on("end", function() {
      server.emit("close", proxyRes, proxySocket, proxyHead);
    });
    socket.on("error", function() {
      proxySocket.end();
    });
    setupSocket(proxySocket);
    if (proxyHead && proxyHead.length) {
      proxySocket.unshift(proxyHead);
    }
    if (!req.headers["x-method-ori"]) {
      socket.write("HTTP/1.1 101 Switching Protocols\r\n");
      socket.write(Object.keys(proxyRes.headers).map(function(i) {
        return i + ": " + proxyRes.headers[i];
      }).join("\r\n") + "\r\n\r\n");
    }
    let proxyStream = proxySocket;
    if (options.createWsServerTransformStream) {
      const wsServerTransformStream = options.createWsServerTransformStream(
        req,
        proxyReq,
        proxyRes
      );
      wsServerTransformStream.on("error", onOutgoingError);
      proxyStream = proxyStream.pipe(wsServerTransformStream);
    }
    proxyStream = proxyStream.pipe(socket);
    if (options.createWsClientTransformStream) {
      const wsClientTransformStream = options.createWsClientTransformStream(
        req,
        proxyReq,
        proxyRes
      );
      wsClientTransformStream.on("error", onOutgoingError);
      proxyStream = proxyStream.pipe(wsClientTransformStream);
    }
    proxyStream.pipe(proxySocket);
    server.emit("open", proxySocket, proxyReq, req);
  });
  if (server) {
    let asyncHandler;
    const asyncContext = (callback2) => {
      asyncHandler = callback2;
    };
    server.emit("proxyReqWs", proxyReq, req, socket, options, head, asyncContext);
    if (asyncHandler) {
      await asyncHandler();
    }
  }
  proxyReq.end();
  function onOutgoingError(err) {
    if (callback) {
      callback(err, req, socket);
    } else {
      server.emit("error", err, req, socket);
    }
    socket.end();
  }
}

const wsIncoming = {
    __proto__: null,
    checkMethodAndHeader: checkMethodAndHeader,
    XHeaders: XHeaders,
    stream: stream
};

var __getProtoOf = Object.getPrototypeOf;
var __reflectGet = Reflect.get;
var __accessCheck = (obj, member, msg) => {
  if (!member.has(obj))
    throw TypeError("Cannot " + msg);
};
var __privateAdd = (obj, member, value) => {
  if (member.has(obj))
    throw TypeError("Cannot add the same private member more than once");
  member instanceof WeakSet ? member.add(obj) : member.set(obj, value);
};
var __privateMethod = (obj, member, method) => {
  __accessCheck(obj, member, "access private method");
  return method;
};
var __superGet = (cls, obj, key) => __reflectGet(__getProtoOf(cls), key, obj);
var _createProxyPassthrough, createProxyPassthrough_fn;
const _ProxyServer = class extends events.EventEmitter {
  constructor(options = {}) {
    super();
    __privateAdd(this, _createProxyPassthrough);
    this.$server = void 0;
    options.prependPath = options.prependPath === false ? false : true;
    this.web = __privateMethod(this, _createProxyPassthrough, createProxyPassthrough_fn).call(this, "web")(options);
    this.ws = __privateMethod(this, _createProxyPassthrough, createProxyPassthrough_fn).call(this, "ws")(options);
    this.options = options;
    this.webPasses = Object.keys(webIncoming).map((pass) => webIncoming[pass]);
    this.wsPasses = Object.keys(wsIncoming).map((pass) => wsIncoming[pass]);
    super.on("error", this.onError);
  }
  onError(err) {
    if (super.listeners("error").length === 1) {
      throw err;
    }
  }
  listen(port, hostname) {
    const closure = (req, res) => {
      this.web(req, res);
    };
    this.$server = this.options.ssl ? https__default.createServer(this.options.ssl, closure) : http__default.createServer(closure);
    if (this.options.ws) {
      this.$server.on("upgrade", (req, socket, head) => {
        this.ws(req, socket, head);
      });
    }
    this.$server.listen(port, hostname);
    return this;
  }
  close(callback) {
    if (this.$server) {
      this.$server.close(() => {
        this.$server = void 0;
        callback?.();
      });
    }
  }
  before(type, passName, callback) {
    const passes = type === "ws" ? this.wsPasses : this.webPasses;
    let i = -1;
    passes.forEach(function(v, idx) {
      if (v.name === passName)
        i = idx;
    });
    if (i === -1)
      throw new Error("No such pass");
    passes.splice(i, 0, callback);
  }
  after(type, passName, callback) {
    const passes = type === "ws" ? this.wsPasses : this.webPasses;
    let i = -1;
    passes.forEach(function(v, idx) {
      if (v.name === passName)
        i = idx;
    });
    if (i === -1)
      throw new Error("No such pass");
    passes.splice(i++, 0, callback);
  }
};
let ProxyServer = _ProxyServer;
_createProxyPassthrough = new WeakSet();
createProxyPassthrough_fn = function(type) {
  return (options) => {
    return (req, res, ...args) => {
      const passes = type === "ws" ? this.wsPasses : this.webPasses;
      let index = args.length - 1;
      let head;
      let callback;
      let proxyOptions = options || this.options;
      if (typeof args[index] === "function") {
        callback = args[index];
        index--;
      }
      if (!(args[index] instanceof buffer.Buffer) && args[index] !== res) {
        proxyOptions = Object.assign({}, options);
        Object.assign(proxyOptions, args[index]);
        index--;
      }
      if (args[index] instanceof buffer.Buffer) {
        head = args[index];
      }
      for (const opt of ["target", "forward"]) {
        if (typeof proxyOptions[opt] === "string") {
          proxyOptions[opt] = new URL(proxyOptions[opt]);
        }
      }
      if (!proxyOptions.target && !proxyOptions.forward) {
        return __superGet(_ProxyServer.prototype, this, "emit").call(this, "error", new Error("Must provide a proper URL as target"));
      }
      for (let i = 0; i < passes.length; i++) {
        if (!proxyOptions?.http2 && typeof passes[i] === "function" && passes[i](req, res, proxyOptions, head, this, callback)) {
          break;
        }
      }
    };
  };
};
const createProxyServer = (options) => {
  return new ProxyServer(options);
};
const createServer = (options) => {
  return new ProxyServer(options);
};
const createProxy = (options) => {
  return new ProxyServer(options);
};

exports.ProxyServer = ProxyServer;
exports.createProxy = createProxy;
exports.createProxyServer = createProxyServer;
exports.createServer = createServer;
